--- /dev/null	2012-10-26 17:09:24.000000000 +0200
+++ new/usr/src/uts/common/fs/zfs/far_count.c	2012-10-26 17:09:24.594803000 +0200
@@ -0,0 +1,152 @@
+/*
+ * CDDL HEADER START
+ *
+ * The contents of this file are subject to the terms of the
+ * Common Development and Distribution License (the "License").
+ * You may not use this file except in compliance with the License.
+ *
+ * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
+ * or http://www.opensolaris.org/os/licensing.
+ * See the License for the specific language governing permissions
+ * and limitations under the License.
+ *
+ * When distributing Covered Code, include this CDDL HEADER in each
+ * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
+ * If applicable, add the following below this CDDL HEADER, with the
+ * fields enclosed by brackets "[]" replaced with your own identifying
+ * information: Portions Copyright [yyyy] [name of copyright owner]
+ *
+ * CDDL HEADER END
+ */
+/*
+ * Copyright (c) 2012 STRATO AG. All rights reserved.
+ */
+#include <sys/zfs_context.h>
+#include <sys/errno.h>
+#include <sys/far_impl.h>
+
+static void
+far_count_value_free(mod_hash_val_t val)
+{
+	far_count_elem_t *fce = val;
+	kmem_free(fce, sizeof (*fce));
+}
+
+static void
+far_count_key_free(mod_hash_key_t key)
+{
+	kmem_free(key, sizeof (uint64_t));
+}
+
+static int
+far_count_cmp(const void *aa, const void *bb)
+{
+	const far_count_elem_t *a = aa;
+	const far_count_elem_t *b = bb;
+
+	if (a->fce_ino < b->fce_ino)
+		return -1;
+	if (a->fce_ino > b->fce_ino)
+		return 1;
+	return 0;
+}
+
+int
+far_add_count(far_counter_t *fc, uint64_t ino, uint64_t inc,
+    uint64_t aux, uint64_t *new_count, uint64_t *old_aux)
+{
+	far_count_elem_t *fce;
+	far_count_elem_t e = { .fce_ino = ino };
+	avl_index_t where;
+
+	fce = avl_find(&fc->fc_avl, &e, &where);
+
+	if (!fce) {
+		fce = kmem_alloc(sizeof (*fce), KM_SLEEP);
+		fce->fce_count = 0;
+		fce->fce_aux = 0;
+		fce->fce_ino = ino;
+		avl_insert(&fc->fc_avl, fce, where);
+	}
+
+	if (old_aux) {
+		*old_aux = fce->fce_aux;
+		fce->fce_aux = aux;
+	}
+	fce->fce_count += inc;
+
+	if (new_count)
+		*new_count = fce->fce_count;
+
+	return (0);
+}
+
+int
+far_get_count(far_counter_t *fc, uint64_t ino, uint64_t *count, uint64_t *aux)
+{
+	far_count_elem_t *fce;
+	far_count_elem_t e = { .fce_ino = ino };
+
+	fce = avl_find(&fc->fc_avl, &e, NULL);
+	if (!fce) {
+		if (count)
+			*count = 0;
+		if (aux)
+			*aux = 0;
+		return (ENOENT);
+	} else {
+		if (count)
+			*count = fce->fce_count;
+		if (aux)
+			*aux = fce->fce_aux;
+		return (0);
+	}
+}
+
+void
+far_free_count(far_counter_t *fc, uint64_t ino)
+{
+	far_count_elem_t *fce;
+	far_count_elem_t e = { .fce_ino = ino };
+
+	fce = avl_find(&fc->fc_avl, &e, NULL);
+	if (!fce)
+		return;
+	avl_remove(&fc->fc_avl, fce);
+	kmem_free(fce, sizeof (*fce));
+}
+
+int
+far_count_init(far_counter_t *fc, char *name)
+{
+	avl_create(&fc->fc_avl, far_count_cmp, sizeof (far_count_elem_t),
+	    offsetof(far_count_elem_t, fce_avl_node));
+	fc->fc_name = name;
+
+	return 0;
+}
+
+int
+far_count_fini(far_counter_t *fc)
+{
+	far_count_elem_t *fce;
+	int ret = 0;
+
+	while ((fce = avl_first(&fc->fc_avl))) {
+		/*
+		 * a count of zero might be left over if a file had > 1 links
+		 * and be replaced by a file with > 1 link. see test 041.034
+		 */
+		if (fce->fce_count != 0) {
+			cmn_err(CE_NOTE, "far_assert_count_empty: %s ino %"
+			    PRIu64 " count %" PRIu64"\n", fc->fc_name,
+			    fce->fce_ino, fce->fce_count);
+			++ret;
+		}
+		avl_remove(&fc->fc_avl, fce);
+		kmem_free(fce, sizeof (*fce));
+	}
+	avl_destroy(&fc->fc_avl);
+
+	return (ret);
+}
